@@ -43,14 +43,15 @@
 
 #include "include/gpfs.h"
 
-/* credential lifetime (1h) */
-uint32_t CredentialLifetime = 3600;
-
-/* static filesystem info.
- * The access is thread-safe because
- * it is read-only, except during initialization.
- */
-struct fsal_staticfsinfo_t global_fs_info;
+#define FSAL_INTERNAL_ERROR(__error, __msg)				      \
+({									      \
+	if ((__error) == EUNATCH)					      \
+		LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");	      \
+									      \
+	LogFullDebug(COMPONENT_FSAL, "%s returned errno=%d", __msg, __error); \
+									      \
+	fsalstat(posix2fsal_error(__error), __error);			      \
+})
 
 /*********************************************************************
  *
@@ -58,35 +59,6 @@ struct fsal_staticfsinfo_t global_fs_info;
  *
  ********************************************************************/
 
-/**
- *  @brief Open a file by handle within an export.
- *
- *  @param dirfd Descriptor
- *  @param gpfs_fh Opaque filehandle
- *  @param fd File descriptor openned by the function
- *  @param oflags Flags to open the file with
- *  @param reopen Bool specifying whether a reopen is wanted
- *
- *  @return status of operation
- */
-fsal_status_t
-fsal_internal_handle2fd(int dirfd, struct gpfs_file_handle *gpfs_fh,
-			int *pfd, int oflags, bool reopen)
-{
-	fsal_status_t status;
-
-	if (!gpfs_fh || !pfd)
-		return fsalstat(ERR_FSAL_FAULT, 0);
-
-	status = fsal_internal_handle2fd_at(dirfd, gpfs_fh, pfd, oflags,
-					    reopen);
-
-	if (FSAL_IS_ERROR(status))
-		return status;
-
-	return fsalstat(ERR_FSAL_NO_ERROR, 0);
-}
-
 /**
  *  @brief Close by fd
  *
@@ -96,26 +68,15 @@ fsal_internal_handle2fd(int dirfd, struct gpfs_file_handle *gpfs_fh,
  */
 fsal_status_t fsal_internal_close(int fd, void *owner, int cflags)
 {
-	int rc = 0;
 	struct close_file_arg carg;
-	int errsv = 0;
 
 	carg.mountdirfd = fd;
 	carg.close_fd = fd;
 	carg.close_flags = cflags;
 	carg.close_owner = owner;
 
-	rc = gpfs_ganesha(OPENHANDLE_CLOSE_FILE, &carg);
-	errsv = errno;
-
-	LogFullDebug(COMPONENT_FSAL, "OPENHANDLE_CLOSE_FILE returned: rc %d",
-		     rc);
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_CLOSE_FILE, &carg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_CLOSE_FILE");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -127,64 +88,28 @@ fsal_status_t fsal_internal_close(int fd, void *owner, int cflags)
  *  @param gpfs_fh Opaque filehandle
  *  @param fd File descriptor openned by the function
  *  @param oflags Flags to open the file with
- *  @param reopen Bool specifying whether a reopen is wanted
  *
  *  @return status of operation
  */
 fsal_status_t
-fsal_internal_handle2fd_at(int dirfd, struct gpfs_file_handle *gpfs_fh,
-			   int *fd, int oflags, bool reopen)
+fsal_internal_handle2fd(int dirfd, struct gpfs_file_handle *gpfs_fh,
+			int *fd, int oflags)
 {
-	int rc = 0;
-	int errsv = 0;
-	union {
-		struct open_arg oarg;
-		struct open_share_arg sarg;
-	} u;
-
-	if (!gpfs_fh || !fd)
-		return fsalstat(ERR_FSAL_FAULT, 0);
+	struct open_arg oarg = {0};
+	int rc;
 
 	if (op_ctx && op_ctx->client && op_ctx->client->hostaddr_str)
-		u.oarg.cli_ip = op_ctx->client->hostaddr_str;
-	else
-		u.oarg.cli_ip = NULL;
+		oarg.cli_ip = op_ctx->client->hostaddr_str;
 
-	if (reopen) {
-		u.sarg.mountdirfd = dirfd;
-		u.sarg.handle = gpfs_fh;
-		u.sarg.flags = oflags;
-		u.sarg.openfd = *fd;
-		/* share_access and share_deny are unused by REOPEN */
-		u.sarg.share_access = 0;
-		u.sarg.share_deny = 0;
-		rc = gpfs_ganesha(OPENHANDLE_REOPEN_BY_FD, &u.sarg);
-		errsv = errno;
-		LogFullDebug(COMPONENT_FSAL,
-			     "OPENHANDLE_REOPEN_BY_FD returned: rc %d", rc);
-	} else {
-		u.oarg.mountdirfd = dirfd;
-		u.oarg.handle = gpfs_fh;
-		u.oarg.flags = oflags;
+	oarg.mountdirfd = dirfd;
+	oarg.handle = gpfs_fh;
+	oarg.flags = oflags;
 
-		rc = gpfs_ganesha(OPENHANDLE_OPEN_BY_HANDLE, &u.oarg);
-		errsv = errno;
-		LogFullDebug(COMPONENT_FSAL,
-			     "OPENHANDLE_OPEN_BY_HANDLE returned: rc %d", rc);
-	}
+	rc = gpfs_ganesha(OPENHANDLE_OPEN_BY_HANDLE, &oarg);
+	if (unlikely(rc < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_OPEN_BY_HANDLE");
 
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
-
-	/* gpfs_open returns fd number for OPENHANDLE_OPEN_BY_HANDLE,
-	 * but only returns 0 for success for OPENHANDLE_REOPEN_BY_FD
-	 * operation. We already have correct (*fd) in reopen case!
-	 */
-	if (!reopen)
-		*fd = rc;
+	*fd = rc;
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -201,11 +126,9 @@ fsal_internal_handle2fd_at(int dirfd, struct gpfs_file_handle *gpfs_fh,
 fsal_status_t
 fsal_internal_get_handle_at(int dfd, const char *fs_name,
 			    struct gpfs_file_handle *gpfs_fh,
-			    int expfd, int *expfdP)
+			    int expfd)
 {
-	int rc;
 	struct name_handle_arg harg;
-	int errsv = 0;
 
 	if (!gpfs_fh)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -219,20 +142,12 @@ fsal_internal_get_handle_at(int dfd, const char *fs_name,
 	harg.expfd = expfd;
 	harg.flag = 0;
 
-	LogFullDebug(COMPONENT_FSAL, "Lookup handle at for %d %s", dfd,
+	LogFullDebug(COMPONENT_FSAL, "Lookup handle at %d for %s", dfd,
 		     fs_name);
 
-	rc = gpfs_ganesha(OPENHANDLE_NAME_TO_HANDLE, &harg);
-	errsv = errno;
+	if (unlikely(gpfs_ganesha(OPENHANDLE_NAME_TO_HANDLE, &harg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_NAME_TO_HANDLE");
 
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
-	LogFullDebug(COMPONENT_FSAL, "Lookup fd %d for %s", rc, fs_name);
-	if (expfdP)
-		*expfdP = rc;
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
@@ -250,9 +165,7 @@ fsal_status_t
 fsal_internal_get_fh(int dirfd, struct gpfs_file_handle *gpfs_fh,
 		     const char *fs_name, struct gpfs_file_handle *gpfs_fh_out)
 {
-	int rc;
 	struct get_handle_arg harg;
-	int errsv = 0;
 
 	if (!gpfs_fh_out || !gpfs_fh || !fs_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -268,14 +181,8 @@ fsal_internal_get_fh(int dirfd, struct gpfs_file_handle *gpfs_fh,
 
 	LogFullDebug(COMPONENT_FSAL, "Lookup handle for %s", fs_name);
 
-	rc = gpfs_ganesha(OPENHANDLE_GET_HANDLE, &harg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_GET_HANDLE, &harg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_GET_HANDLE");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -289,11 +196,9 @@ fsal_internal_get_fh(int dirfd, struct gpfs_file_handle *gpfs_fh,
  *  @return status of operation
  */
 fsal_status_t
-fsal_internal_fd2handle(int fd, struct gpfs_file_handle *gpfs_fh, int *expfdP)
+fsal_internal_fd2handle(int fd, struct gpfs_file_handle *gpfs_fh)
 {
-	int rc;
-	struct name_handle_arg harg;
-	int errsv = 0;
+	struct name_handle_arg harg = {0};
 
 	if (!gpfs_fh)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -302,27 +207,12 @@ fsal_internal_fd2handle(int fd, struct gpfs_file_handle *gpfs_fh, int *expfdP)
 	harg.handle->handle_size = GPFS_MAX_FH_SIZE;
 	harg.handle->handle_key_size = OPENHANDLE_KEY_LEN;
 	harg.handle->handle_version = OPENHANDLE_VERSION;
-	harg.name = NULL;
 	harg.dfd = fd;
-	harg.flag = 0;
-	if (expfdP)
-		harg.expfd = *expfdP;
-	else
-		harg.expfd = 0;
 
 	LogFullDebug(COMPONENT_FSAL, "Lookup handle by fd for %d", fd);
 
-	rc = gpfs_ganesha(OPENHANDLE_NAME_TO_HANDLE, &harg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
-	LogFullDebug(COMPONENT_FSAL, "get expfd in %d out %d", fd, rc);
-	if (expfdP)
-		*expfdP = rc;
+	if (unlikely(gpfs_ganesha(OPENHANDLE_NAME_TO_HANDLE, &harg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_NAME_TO_HANDLE");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -341,9 +231,7 @@ fsal_status_t
 fsal_internal_link_fh(int dirfd, struct gpfs_file_handle *gpfs_fh_tgt,
 		      struct gpfs_file_handle *gpfs_fh, const char *link_name)
 {
-	int rc;
 	struct link_fh_arg linkarg;
-	int errsv = 0;
 
 	if (!link_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -354,14 +242,8 @@ fsal_internal_link_fh(int dirfd, struct gpfs_file_handle *gpfs_fh_tgt,
 	linkarg.dir_fh = gpfs_fh;
 	linkarg.dst_fh = gpfs_fh_tgt;
 
-	rc = gpfs_ganesha(OPENHANDLE_LINK_BY_FH, &linkarg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_LINK_BY_FH, &linkarg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_LINK_BY_FH");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 
@@ -381,9 +263,7 @@ fsal_status_t
 fsal_internal_stat_name(int dirfd, struct gpfs_file_handle *gpfs_fh,
 			const char *stat_name, struct stat *buf)
 {
-	int rc;
 	struct stat_name_arg statarg;
-	int errsv = 0;
 
 	if (!stat_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -394,14 +274,8 @@ fsal_internal_stat_name(int dirfd, struct gpfs_file_handle *gpfs_fh,
 	statarg.handle = gpfs_fh;
 	statarg.buf = buf;
 
-	rc = gpfs_ganesha(OPENHANDLE_STAT_BY_NAME, &statarg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_STAT_BY_NAME, &statarg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_STAT_BY_NAME");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -420,9 +294,7 @@ fsal_status_t
 fsal_internal_unlink(int dirfd, struct gpfs_file_handle *gpfs_fh,
 		     const char *stat_name, struct stat *buf)
 {
-	int rc;
 	struct stat_name_arg statarg;
-	int errsv = 0;
 
 	if (!stat_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -433,14 +305,8 @@ fsal_internal_unlink(int dirfd, struct gpfs_file_handle *gpfs_fh,
 	statarg.handle = gpfs_fh;
 	statarg.buf = buf;
 
-	rc = gpfs_ganesha(OPENHANDLE_UNLINK_BY_NAME, &statarg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_UNLINK_BY_NAME, &statarg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_UNLINK_BY_NAME");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -462,39 +328,29 @@ fsal_internal_create(struct fsal_obj_handle *dir_hdl, const char *stat_name,
 		     mode_t mode, int posix_flags, struct gpfs_file_handle *fh,
 		     struct stat *buf)
 {
-	int rc;
-	struct create_name_arg crarg;
-	struct gpfs_filesystem *gpfs_fs;
-	struct gpfs_fsal_obj_handle *gpfs_hdl;
-	int errsv = 0;
+	struct create_name_arg crarg = {0};
+	struct gpfs_fsal_export *exp = container_of(op_ctx->fsal_export,
+					struct gpfs_fsal_export, export);
+	int export_fd = exp->export_fd;
 
 	if (!stat_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
 
-	gpfs_hdl =
-	   container_of(dir_hdl, struct gpfs_fsal_obj_handle, obj_handle);
-	gpfs_fs = dir_hdl->fs->private_data;
-
-	crarg.mountdirfd = gpfs_fs->root_fd;
+	crarg.mountdirfd = export_fd;
 	crarg.mode = mode;
 	crarg.dev = posix_flags;
 	crarg.len = strlen(stat_name);
 	crarg.name = stat_name;
-	crarg.dir_fh = gpfs_hdl->handle;
 	crarg.new_fh = fh;
 	crarg.new_fh->handle_size = GPFS_MAX_FH_SIZE;
 	crarg.new_fh->handle_key_size = OPENHANDLE_KEY_LEN;
 	crarg.new_fh->handle_version = OPENHANDLE_VERSION;
 	crarg.buf = buf;
+	crarg.dir_fh = container_of(dir_hdl, struct gpfs_fsal_obj_handle,
+				    obj_handle)->handle;
 
-	rc = gpfs_ganesha(OPENHANDLE_CREATE_BY_NAME, &crarg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_CREATE_BY_NAME, &crarg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_CREATE_BY_NAME");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -504,39 +360,29 @@ fsal_internal_mknode(struct fsal_obj_handle *dir_hdl, const char *stat_name,
 		     mode_t mode, dev_t dev, struct gpfs_file_handle *fh,
 		     struct stat *buf)
 {
-	int rc;
-	struct create_name_arg crarg;
-	struct gpfs_filesystem *gpfs_fs;
-	struct gpfs_fsal_obj_handle *gpfs_hdl;
-	int errsv = 0;
+	struct create_name_arg crarg = {0};
+	struct gpfs_fsal_export *exp = container_of(op_ctx->fsal_export,
+					struct gpfs_fsal_export, export);
+	int export_fd = exp->export_fd;
 
 	if (!stat_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
 
-	gpfs_hdl =
-	    container_of(dir_hdl, struct gpfs_fsal_obj_handle, obj_handle);
-	gpfs_fs = dir_hdl->fs->private_data;
-
-	crarg.mountdirfd = gpfs_fs->root_fd;
+	crarg.mountdirfd = export_fd;
 	crarg.mode = mode;
 	crarg.dev = dev;
 	crarg.len = strlen(stat_name);
 	crarg.name = stat_name;
-	crarg.dir_fh = gpfs_hdl->handle;
 	crarg.new_fh = fh;
 	crarg.new_fh->handle_size = GPFS_MAX_FH_SIZE;
 	crarg.new_fh->handle_key_size = OPENHANDLE_KEY_LEN;
 	crarg.new_fh->handle_version = OPENHANDLE_VERSION;
 	crarg.buf = buf;
+	crarg.dir_fh = container_of(dir_hdl, struct gpfs_fsal_obj_handle,
+				    obj_handle)->handle;
 
-	rc = gpfs_ganesha(OPENHANDLE_MKNODE_BY_NAME, &crarg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_MKNODE_BY_NAME, &crarg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_MKNODE_BY_NAME");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -557,9 +403,7 @@ fsal_internal_rename_fh(int dirfd, struct gpfs_file_handle *gpfs_fh_old,
 			struct gpfs_file_handle *gpfs_fh_new,
 			const char *old_name, const char *new_name)
 {
-	int rc;
 	struct rename_fh_arg renamearg;
-	int errsv = 0;
 
 	if (!old_name || !new_name)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -572,14 +416,8 @@ fsal_internal_rename_fh(int dirfd, struct gpfs_file_handle *gpfs_fh_old,
 	renamearg.old_fh = gpfs_fh_old;
 	renamearg.new_fh = gpfs_fh_new;
 
-	rc = gpfs_ganesha(OPENHANDLE_RENAME_BY_FH, &renamearg);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(gpfs_ganesha(OPENHANDLE_RENAME_BY_FH, &renamearg) < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_RENAME_BY_FH");
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
@@ -596,30 +434,23 @@ fsal_internal_rename_fh(int dirfd, struct gpfs_file_handle *gpfs_fh_old,
  */
 fsal_status_t
 fsal_readlink_by_handle(int dirfd, struct gpfs_file_handle *gpfs_fh, char *buf,
-			size_t *maxlen)
+			size_t maxlen)
 {
-	int rc;
 	struct readlink_fh_arg readlinkarg;
-	int errsv = 0;
+	int rc;
 
 	readlinkarg.mountdirfd = dirfd;
 	readlinkarg.handle = gpfs_fh;
 	readlinkarg.buffer = buf;
-	readlinkarg.size = *maxlen;
+	readlinkarg.size = maxlen - 1; /* reserve 1 for terminating 0-byte */
 
 	rc = gpfs_ganesha(OPENHANDLE_READLINK_BY_FH, &readlinkarg);
-	errsv = errno;
 
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		return fsalstat(posix2fsal_error(errsv), errsv);
-	}
+	if (unlikely(rc < 0))
+		return FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_READLINK_BY_FH");
+
+	buf[rc] = '\0';
 
-	if (rc < *maxlen) {
-		buf[rc] = '\0';
-		*maxlen = rc;
-	}
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
@@ -631,17 +462,10 @@ fsal_readlink_by_handle(int dirfd, struct gpfs_file_handle *gpfs_fh, char *buf,
 int fsal_internal_version(void)
 {
 	int rc;
-	int errsv = 0;
 
-	rc = gpfs_ganesha(OPENHANDLE_GET_VERSION2, &rc);
-	errsv = errno;
-
-	if (rc < 0) {
-		if (errsv == EUNATCH)
-			LogFatal(COMPONENT_FSAL, "GPFS Returned EUNATCH");
-		LogDebug(COMPONENT_FSAL, "GPFS get version failed with rc %d",
-			 rc);
-	} else
+	if (unlikely(gpfs_ganesha(OPENHANDLE_GET_VERSION2, &rc) < 0))
+		FSAL_INTERNAL_ERROR(errno, "OPENHANDLE_GET_VERSION2");
+	else
 		LogDebug(COMPONENT_FSAL, "GPFS get version %d", rc);
 
 	return rc;
@@ -660,26 +484,24 @@ int fsal_internal_version(void)
  */
 fsal_status_t
 fsal_get_xstat_by_handle(int dirfd, struct gpfs_file_handle *gpfs_fh,
-			 gpfsfsal_xstat_t *buffxstat,
-			 uint32_t *expire_time_attr, bool expire, bool use_acl)
+			 gpfsfsal_xstat_t *buffxstat, gpfs_acl_t *acl_buf,
+			 unsigned int acl_buflen, uint32_t *expire_time_attr,
+			 bool expire, bool use_acl)
 {
+	struct xstat_arg xstatarg = {0};
+	int errsv;
 	int rc;
-	struct xstat_arg xstatarg;
-	int errsv = 0;
 
 	if (!gpfs_fh || !buffxstat)
 		return fsalstat(ERR_FSAL_FAULT, 0);
 
 	/* Initialize acl header so that GPFS knows what we want. */
 	if (use_acl) {
-		gpfs_acl_t *pacl_gpfs;
-
-		pacl_gpfs = (gpfs_acl_t *) buffxstat->buffacl;
-		pacl_gpfs->acl_level = 0;
-		pacl_gpfs->acl_version = GPFS_ACL_VERSION_NFS4;
-		pacl_gpfs->acl_type = GPFS_ACL_TYPE_NFS4;
-		pacl_gpfs->acl_len = GPFS_ACL_BUF_SIZE;
-		xstatarg.acl = pacl_gpfs;
+		acl_buf->acl_level = 0;
+		acl_buf->acl_version = GPFS_ACL_VERSION_NFS4;
+		acl_buf->acl_type = GPFS_ACL_TYPE_NFS4;
+		acl_buf->acl_len = acl_buflen;
+		xstatarg.acl = acl_buf;
 		xstatarg.attr_valid = XATTR_STAT | XATTR_ACL;
 	} else {
 		xstatarg.acl = NULL;
@@ -703,7 +525,8 @@ fsal_get_xstat_by_handle(int dirfd, struct gpfs_file_handle *gpfs_fh,
 		 dirfd, rc, gpfs_fh->handle_size);
 
 	if (rc < 0) {
-		if (errsv == ENODATA) {
+		switch (errsv) {
+		case ENODATA:
 			/* For the special file that do not have ACL, GPFS
 			   returns ENODATA. In this case, return okay with
 			   stat.
@@ -712,7 +535,28 @@ fsal_get_xstat_by_handle(int dirfd, struct gpfs_file_handle *gpfs_fh,
 			LogFullDebug(COMPONENT_FSAL,
 				     "retrieved only stat, not acl");
 			return fsalstat(ERR_FSAL_NO_ERROR, 0);
-		} else {
+
+		case ENOSPC:
+			/* If the supplied acl buffer is too small, we
+			 * get this errno! acl_len will be updated to
+			 * the required length.
+			 *
+			 * Return success and let the caller check the length
+			 */
+			if (use_acl && acl_buf->acl_len > acl_buflen) {
+				LogFullDebug(COMPONENT_FSAL,
+					"fsal_get_xstat_by_handle returned buffer too small, passed len: %u, required len: %u, ",
+					acl_buflen, acl_buf->acl_len);
+				errno = 0;
+				break;
+			}
+
+			LogWarn(COMPONENT_FSAL,
+				"fsal_get_xstat_by_handle returned bogus ENOSPC, passed len: %u, required len: %u",
+				acl_buflen, acl_buf->acl_len);
+			return fsalstat(ERR_FSAL_SERVERFAULT, errsv);
+
+		default:
 			/* Handle other errors. */
 			LogFullDebug(COMPONENT_FSAL,
 				     "fsal_get_xstat_by_handle returned errno:%d -- %s",
@@ -747,10 +591,12 @@ fsal_get_xstat_by_handle(int dirfd, struct gpfs_file_handle *gpfs_fh,
 fsal_status_t
 fsal_set_xstat_by_handle(int dirfd, const struct req_op_context *op_ctx,
 			 struct gpfs_file_handle *gpfs_fh, int attr_valid,
-			 int attr_changed, gpfsfsal_xstat_t *buffxstat)
+			 int attr_changed, gpfsfsal_xstat_t *buffxstat,
+			 gpfs_acl_t *acl_buf)
 {
-	int rc, errsv;
-	struct xstat_arg xstatarg;
+	struct xstat_arg xstatarg = {0};
+	int errsv;
+	int rc;
 
 	if (!gpfs_fh || !buffxstat)
 		return fsalstat(ERR_FSAL_FAULT, 0);
@@ -758,7 +604,7 @@ fsal_set_xstat_by_handle(int dirfd, const struct req_op_context *op_ctx,
 	xstatarg.attr_valid = attr_valid;
 	xstatarg.mountdirfd = dirfd;
 	xstatarg.handle = gpfs_fh;
-	xstatarg.acl = (gpfs_acl_t *) buffxstat->buffacl;
+	xstatarg.acl = acl_buf;
 	xstatarg.attr_changed = attr_changed;
 	xstatarg.buf = &buffxstat->buffstat;
 
@@ -798,19 +644,15 @@ fsal_status_t
 fsal_trucate_by_handle(int dirfd, const struct req_op_context *op_ctx,
 		       struct gpfs_file_handle *gpfs_fh, u_int64_t size)
 {
-	int attr_valid;
-	int attr_changed;
-	gpfsfsal_xstat_t buffxstat;
+	gpfsfsal_xstat_t buffxstat = {0};
 
 	if (!gpfs_fh || !op_ctx)
 		return fsalstat(ERR_FSAL_FAULT, 0);
 
-	attr_valid = XATTR_STAT;
-	attr_changed = XATTR_SIZE;
 	buffxstat.buffstat.st_size = size;
 
-	return fsal_set_xstat_by_handle(dirfd, op_ctx, gpfs_fh, attr_valid,
-					attr_changed, &buffxstat);
+	return fsal_set_xstat_by_handle(dirfd, op_ctx, gpfs_fh, XATTR_STAT,
+					XATTR_SIZE, &buffxstat, NULL);
 }
 
 /**
