@@ -47,6 +47,9 @@
 #include "vfs_methods.h"
 #include <os/subr.h>
 #include "subfsal.h"
+#include "city.h"
+#include "nfs_core.h"
+#include "nfs_proto_tools.h"
 
 /* helpers
  */
@@ -95,6 +98,9 @@ struct vfs_fsal_obj_handle *alloc_handle(int dirfd,
 	if (hdl->obj_handle.type == REGULAR_FILE) {
 		hdl->u.file.fd.fd = -1;	/* no open on this yet */
 		hdl->u.file.fd.openflags = FSAL_O_CLOSED;
+	} else if (hdl->obj_handle.type == DIRECTORY) {
+		hdl->u.directory.path = NULL;
+		hdl->u.directory.fs_location = NULL;
 	} else if (hdl->obj_handle.type == SYMBOLIC_LINK) {
 		ssize_t retlink;
 		size_t len = stat->st_size + 1;
@@ -154,57 +160,23 @@ struct vfs_fsal_obj_handle *alloc_handle(int dirfd,
 	return NULL;
 }
 
-/* handle methods
- */
-
-/* lookup
- * deprecated NULL parent && NULL path implies root handle
- */
-
-static fsal_status_t lookup(struct fsal_obj_handle *parent,
-			    const char *path, struct fsal_obj_handle **handle,
-			    struct attrlist *attrs_out)
+static fsal_status_t lookup_with_fd(struct vfs_fsal_obj_handle *parent_hdl,
+				    int dirfd, const char *path,
+				    struct fsal_obj_handle **handle,
+				    struct attrlist *attrs_out)
 {
-	struct vfs_fsal_obj_handle *parent_hdl, *hdl;
-	fsal_errors_t fsal_error = ERR_FSAL_NO_ERROR;
-	int retval, dirfd;
+	struct vfs_fsal_obj_handle *hdl;
+	int retval, fd;
 	struct stat stat;
 	vfs_file_handle_t *fh = NULL;
 	fsal_dev_t dev;
 	struct fsal_filesystem *fs;
 	bool xfsal = false;
 	fsal_status_t status;
+	fsal_errors_t fsal_error = ERR_FSAL_NO_ERROR;
 
 	vfs_alloc_handle(fh);
 
-	*handle = NULL;		/* poison it first */
-	parent_hdl =
-	    container_of(parent, struct vfs_fsal_obj_handle, obj_handle);
-	if (!parent->obj_ops.handle_is(parent, DIRECTORY)) {
-		LogCrit(COMPONENT_FSAL,
-			"Parent handle is not a directory. hdl = 0x%p", parent);
-		return fsalstat(ERR_FSAL_NOTDIR, 0);
-	}
-
-	if (parent->fsal != parent->fs->fsal) {
-		LogDebug(COMPONENT_FSAL,
-			 "FSAL %s operation for handle belonging to FSAL %s, return EXDEV",
-			 parent->fsal->name,
-			 parent->fs->fsal != NULL
-				? parent->fs->fsal->name
-				: "(none)");
-		return fsalstat(ERR_FSAL_XDEV, EXDEV);
-	}
-
-	fs = parent->fs;
-	dirfd = vfs_fsal_open(parent_hdl, O_PATH | O_NOACCESS, &fsal_error);
-
-	if (dirfd < 0) {
-		LogDebug(COMPONENT_FSAL, "Failed to open parent: %s",
-			 msg_fsal_err(fsal_error));
-		return fsalstat(fsal_error, -dirfd);
-	}
-
 	retval = fstatat(dirfd, path, &stat, AT_SYMLINK_NOFOLLOW);
 
 	if (retval < 0) {
@@ -212,11 +184,12 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 		LogDebug(COMPONENT_FSAL, "Failed to open stat %s: %s", path,
 			 msg_fsal_err(posix2fsal_error(retval)));
 		status = posix2fsal_status(retval);
-		goto direrr;
+		return status;
 	}
 
 	dev = posix2fsal_devt(stat.st_dev);
 
+	fs = parent_hdl->obj_handle.fs;
 	if ((dev.minor != parent_hdl->dev.minor) ||
 	    (dev.major != parent_hdl->dev.major)) {
 		/* XDEV */
@@ -227,10 +200,10 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 				 PRIu64".%"PRIu64,
 				 path, dev.major, dev.minor);
 			status = fsalstat(ERR_FSAL_XDEV, EXDEV);
-			goto direrr;
+			return status;
 		}
 
-		if (fs->fsal != parent->fsal) {
+		if (fs->fsal != parent_hdl->obj_handle.fsal) {
 			xfsal = true;
 			LogDebug(COMPONENT_FSAL,
 				 "Lookup of %s crosses filesystem boundary to file system %s into FSAL %s",
@@ -251,7 +224,7 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 		     (retval == EOPNOTSUPP) ||
 		     (retval == ENOTSUP) ||
 		     xfsal) &&
-		    (fs != parent->fs)) {
+		    (fs != parent_hdl->obj_handle.fs)) {
 			/* Crossed device into territory not handled by
 			 * this FSAL (XFS or VFS). Need to invent a handle.
 			 * The made up handle will be JUST the fsid, we
@@ -262,7 +235,7 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 			LogDebug(COMPONENT_FSAL,
 				 "vfs_name_to_handle %s, inventing FSAL %s handle for FSAL %s filesystem %s",
 				 xfsal ? "skipped" : "failed",
-				 parent->fsal->name,
+				 parent_hdl->obj_handle.fsal->name,
 				 fs->fsal != NULL
 					? fs->fsal->name
 					: "(none)",
@@ -273,14 +246,14 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 			if (retval < 0) {
 				retval = errno;
 				status = posix2fsal_status(retval);
-				goto direrr;
+				return status;
 			}
 
 			retval = 0;
 		} else {
 			/* Some other error */
 			status = posix2fsal_status(retval);
-			goto direrr;
+			return status;
 		}
 	}
 
@@ -290,22 +263,171 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 
 	if (hdl == NULL) {
 		status = fsalstat(ERR_FSAL_NOMEM, ENOMEM);
-		goto direrr;
+		return status;
 	}
 
-	close(dirfd);
-
 	if (attrs_out != NULL) {
-		posix2fsal_attributes(&stat, attrs_out);
+		posix2fsal_attributes_all(&stat, attrs_out);
+	}
 
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
+	/* if it is a directory and the sticky bit is set
+	 * let's look for referral information
+	 */
+
+	if (hdl->obj_handle.type == DIRECTORY &&
+	    attrs_out != NULL &&
+	    is_sticky_bit_set(&hdl->obj_handle, attrs_out) &&
+	    hdl->obj_handle.fs->private_data != NULL) {
+
+		caddr_t xattr_content;
+		size_t attrsize = 0;
+		char proclnk[MAXPATHLEN];
+		char path[MAXPATHLEN];
+		char *spath, *fspath;
+		ssize_t r;
+		uint64 hash;
+		fsal_status_t st;
+
+		struct vfs_filesystem *vfs_fs =
+			hdl->obj_handle.fs->private_data;
+
+		/* the real path of the referral directory is needed.
+		 * it get's stored in u.directory.path
+		 */
+
+		fd = vfs_fsal_open(hdl, O_DIRECTORY, &fsal_error);
+		if (fd < 0) {
+			return fsalstat(fsal_error, -fd);
+		}
+
+		sprintf(proclnk, "/proc/self/fd/%d", fd);
+		r = readlink(proclnk, path, MAXPATHLEN - 1);
+		if (r < 0) {
+			fsal_error = posix2fsal_error(errno);
+			r = errno;
+			LogEvent(COMPONENT_FSAL, "failed to readlink");
+			close(fd);
+			return fsalstat(fsal_error, r);
+		}
+		path[r] = '\0';
+		LogDebug(COMPONENT_FSAL, "fd -> path: %d -> %s",
+			 fd, path);
+		if (hdl->u.directory.path != NULL) {
+			LogFullDebug(COMPONENT_FSAL,
+				     "freeing old directory.path: %s",
+				     hdl->u.directory.path);
+			gsh_free(hdl->u.directory.path);
+		}
+
+		fspath = vfs_fs->fs->path;
+
+		spath = path;
+		if (strncmp(path, fspath, strlen(fspath)) == 0) {
+			spath += strlen(fspath);
+		}
+		hdl->u.directory.path = gsh_strdup(spath);
+
+		/* referral configuration is in a xattr "user.fs_location"
+		 * on the directory in the form
+		 * server:/path/to/referred/directory.
+		 * It gets storeded in u.directory.fs_location
+		 */
+
+		xattr_content = gsh_calloc(XATTR_BUFFERSIZE, sizeof(char));
+
+		st = vfs_getextattr_value_by_name((struct fsal_obj_handle *)hdl,
+						  "user.fs_location",
+						  xattr_content,
+						  XATTR_BUFFERSIZE,
+						  &attrsize);
+
+		if (!FSAL_IS_ERROR(st)) {
+			LogDebug(COMPONENT_FSAL, "user.fs_location: %s",
+				 xattr_content);
+			if (hdl->u.directory.fs_location != NULL) {
+				LogFullDebug(COMPONENT_FSAL,
+					     "freeing old directory.fs_location: %s",
+					     hdl->u.directory.fs_location);
+				gsh_free(hdl->u.directory.fs_location);
+			}
+			hdl->u.directory.fs_location =
+				gsh_strdup(xattr_content);
+
+			/* on a referral the filesystem id has to change
+			 * it get's calculated via a hash from the referral
+			 * and stored in the fsid object of the fsal_obj_handle
+			 */
+
+			hash = CityHash64(xattr_content, attrsize);
+			hdl->obj_handle.fsid.major = hash;
+			hdl->obj_handle.fsid.minor = hash;
+			LogDebug(COMPONENT_NFS_V4,
+				 "fsid.major = %"PRIu64", fsid.minor = %"PRIu64,
+				 hdl->obj_handle.fsid.major,
+				 hdl->obj_handle.fsid.minor);
+		}
+		gsh_free(xattr_content);
+		close(fd);
+
+	} else {
+		/* reset fsid if the sticky bit is not set,
+		 * because a referral was removed
+		 */
+
+		hdl->obj_handle.fsid = hdl->obj_handle.fs->fsid;
 	}
 
 	*handle = &hdl->obj_handle;
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
+}
+
+/* handle methods
+ */
+
+/* lookup
+ * deprecated NULL parent && NULL path implies root handle
+ */
+
+static fsal_status_t lookup(struct fsal_obj_handle *parent,
+			    const char *path, struct fsal_obj_handle **handle,
+			    struct attrlist *attrs_out)
+{
+	struct vfs_fsal_obj_handle *parent_hdl;
+	fsal_errors_t fsal_error = ERR_FSAL_NO_ERROR;
+	int dirfd;
+	fsal_status_t status;
+
+	*handle = NULL;		/* poison it first */
+	parent_hdl =
+	    container_of(parent, struct vfs_fsal_obj_handle, obj_handle);
+	if (!parent->obj_ops.handle_is(parent, DIRECTORY)) {
+		LogCrit(COMPONENT_FSAL,
+			"Parent handle is not a directory. hdl = 0x%p", parent);
+		return fsalstat(ERR_FSAL_NOTDIR, 0);
+	}
+
+	if (parent->fsal != parent->fs->fsal) {
+		LogDebug(COMPONENT_FSAL,
+			 "FSAL %s operation for handle belonging to FSAL %s, return EXDEV",
+			 parent->fsal->name,
+			 parent->fs->fsal != NULL
+				? parent->fs->fsal->name
+				: "(none)");
+		return fsalstat(ERR_FSAL_XDEV, EXDEV);
+	}
+
+	dirfd = vfs_fsal_open(parent_hdl, O_PATH | O_NOACCESS, &fsal_error);
+
+	if (dirfd < 0) {
+		LogDebug(COMPONENT_FSAL, "Failed to open parent: %s",
+			 msg_fsal_err(fsal_error));
+		status = fsalstat(fsal_error, -dirfd);
+		return status;
+	}
+
+	status = lookup_with_fd(parent_hdl, dirfd, path, handle, attrs_out);
+
 
- direrr:
 	close(dirfd);
 	return status;
 }
@@ -441,9 +563,9 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 	*handle = &hdl->obj_handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -460,7 +582,7 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 			status = (*handle)->obj_ops.getattrs(*handle,
 							     attrs_out);
 			if (FSAL_IS_ERROR(status) &&
-			    (attrs_out->mask & ATTR_RDATTR_ERR) == 0) {
+			    (attrs_out->request_mask & ATTR_RDATTR_ERR) == 0) {
 				/* Get attributes failed and caller expected
 				 * to get the attributes.
 				 */
@@ -476,10 +598,7 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 			 * was set on create, just use the stat results we used
 			 * to create the fsal_obj_handle.
 			 */
-			posix2fsal_attributes(&stat, attrs_out);
-
-			/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-			attrs_out->mask &= ~ATTR_RDATTR_ERR;
+			posix2fsal_attributes_all(&stat, attrs_out);
 		}
 	}
 
@@ -499,7 +618,6 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 			      const char *name,
 			      object_file_type_t nodetype,	/* IN */
-			      fsal_dev_t *dev,	/* IN */
 			      struct attrlist *attrib,
 			      struct fsal_obj_handle **handle,
 			      struct attrlist *attrs_out)
@@ -577,11 +695,11 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 	switch (nodetype) {
 	case BLOCK_FILE:
 		unix_mode |= S_IFBLK;
-		unix_dev = makedev(dev->major, dev->minor);
+		unix_dev = makedev(attrib->rawdev.major, attrib->rawdev.minor);
 		break;
 	case CHARACTER_FILE:
 		unix_mode |= S_IFCHR;
-		unix_dev = makedev(dev->major, dev->minor);
+		unix_dev = makedev(attrib->rawdev.major, attrib->rawdev.minor);
 		break;
 	case FIFO_FILE:
 		unix_mode |= S_IFIFO;
@@ -652,9 +770,9 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 	*handle = &hdl->obj_handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -668,7 +786,7 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 			status = (*handle)->obj_ops.getattrs(*handle,
 							     attrs_out);
 			if (FSAL_IS_ERROR(status) &&
-			    (attrs_out->mask & ATTR_RDATTR_ERR) == 0) {
+			    (attrs_out->request_mask & ATTR_RDATTR_ERR) == 0) {
 				/* Get attributes failed and caller expected
 				 * to get the attributes.
 				 */
@@ -684,10 +802,7 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 			 * was set on create, just use the stat results we used
 			 * to create the fsal_obj_handle.
 			 */
-			posix2fsal_attributes(&stat, attrs_out);
-
-			/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-			attrs_out->mask &= ~ATTR_RDATTR_ERR;
+			posix2fsal_attributes_all(&stat, attrs_out);
 		}
 	}
 
@@ -841,9 +956,9 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 	*handle = &hdl->obj_handle;
 
 	/* We handled the mode above. */
-	FSAL_UNSET_MASK(attrib->mask, ATTR_MODE);
+	FSAL_UNSET_MASK(attrib->valid_mask, ATTR_MODE);
 
-	if (attrib->mask) {
+	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
@@ -857,7 +972,7 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 			status = (*handle)->obj_ops.getattrs(*handle,
 							     attrs_out);
 			if (FSAL_IS_ERROR(status) &&
-			    (attrs_out->mask & ATTR_RDATTR_ERR) == 0) {
+			    (attrs_out->request_mask & ATTR_RDATTR_ERR) == 0) {
 				/* Get attributes failed and caller expected
 				 * to get the attributes.
 				 */
@@ -873,10 +988,7 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 			 * was set on create, just use the stat results we used
 			 * to create the fsal_obj_handle.
 			 */
-			posix2fsal_attributes(&stat, attrs_out);
-
-			/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-			attrs_out->mask &= ~ATTR_RDATTR_ERR;
+			posix2fsal_attributes_all(&stat, attrs_out);
 		}
 	}
 
@@ -978,7 +1090,7 @@ static fsal_status_t linkfile(struct fsal_obj_handle *obj_hdl,
 	 * We only take a read lock because we are not changing the state of
 	 * the file descriptor.
 	 */
-	PTHREAD_RWLOCK_rdlock(&obj_hdl->lock);
+	PTHREAD_RWLOCK_rdlock(&obj_hdl->obj_lock);
 
 	if (obj_hdl->type == REGULAR_FILE &&
 	    myself->u.file.fd.openflags != FSAL_O_CLOSED) {
@@ -1036,7 +1148,7 @@ static fsal_status_t linkfile(struct fsal_obj_handle *obj_hdl,
 
  out_unlock:
 
-	PTHREAD_RWLOCK_unlock(&obj_hdl->lock);
+	PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
 
  out:
 	LogFullDebug(COMPONENT_FSAL, "returning %d, %d", fsal_error, retval);
@@ -1088,6 +1200,7 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 	dirfd = vfs_fsal_open(myself, O_RDONLY | O_DIRECTORY, &status.major);
 	if (dirfd < 0) {
 		retval = -dirfd;
+		status = posix2fsal_status(retval);
 		goto out;
 	}
 	seekloc = lseek(dirfd, seekloc, SEEK_SET);
@@ -1110,7 +1223,7 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 		for (bpos = 0; bpos < nread;) {
 			struct fsal_obj_handle *hdl;
 			struct attrlist attrs;
-			bool cb_rc;
+			enum fsal_dir_result cb_rc;
 
 			if (!to_vfs_dirent(buf, bpos, dentryp, baseloc)
 			    || strcmp(dentryp->vd_name, ".") == 0
@@ -1119,11 +1232,12 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 
 			fsal_prepare_attrs(&attrs, attrmask);
 
-			status = lookup(dir_hdl, dentryp->vd_name,
+			status = lookup_with_fd(myself, dirfd, dentryp->vd_name,
 					&hdl, &attrs);
 
-			if (FSAL_IS_ERROR(status))
+			if (FSAL_IS_ERROR(status)) {
 				goto done;
+			}
 
 			/* callback to cache inode */
 			cb_rc = cb(dentryp->vd_name, hdl, &attrs, dir_state,
@@ -1131,7 +1245,8 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 
 			fsal_release_attrs(&attrs);
 
-			if (!cb_rc)
+			/* Read ahead not supported by this FSAL. */
+			if (cb_rc >= DIR_READAHEAD)
 				goto done;
 
  skip:
@@ -1144,7 +1259,6 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 	close(dirfd);
 
  out:
-	status.minor = retval;
 	return status;
 }
 
@@ -1229,7 +1343,7 @@ static fsal_status_t renamefile(struct fsal_obj_handle *obj_hdl,
  * @brief Open file and get attributes.
  *
  * This function opens a file and returns the file descriptor and fetches
- * the attributes. The obj_hdl->lock MUST be held for this call.
+ * the attributes. The obj_hdl->obj_lock MUST be held for this call.
  *
  * @param[in]     exp         The fsal_export the file belongs to
  * @param[in]     myself      File to access
@@ -1407,16 +1521,16 @@ static fsal_status_t file_unlink(struct fsal_obj_handle *dir_hdl,
 	return fsalstat(fsal_error, retval);
 }
 
-/* handle_digest
+/* handle_to_wire
  * fill in the opaque f/s file handle part.
  * we zero the buffer to length first.  This MAY already be done above
  * at which point, remove memset here because the caller is zeroing
  * the whole struct.
  */
 
-static fsal_status_t handle_digest(const struct fsal_obj_handle *obj_hdl,
-				   fsal_digesttype_t output_type,
-				   struct gsh_buffdesc *fh_desc)
+static fsal_status_t handle_to_wire(const struct fsal_obj_handle *obj_hdl,
+				    fsal_digesttype_t output_type,
+				    struct gsh_buffdesc *fh_desc)
 {
 	const struct vfs_fsal_obj_handle *myself;
 
@@ -1490,11 +1604,11 @@ static void release(struct fsal_obj_handle *obj_hdl)
 		/* Take write lock on object to protect file descriptor.
 		 * This can block over an I/O operation.
 		 */
-		PTHREAD_RWLOCK_wrlock(&obj_hdl->lock);
+		PTHREAD_RWLOCK_wrlock(&obj_hdl->obj_lock);
 
 		st = vfs_close_my_fd(&myself->u.file.fd);
 
-		PTHREAD_RWLOCK_unlock(&obj_hdl->lock);
+		PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
 
 		if (FSAL_IS_ERROR(st)) {
 			LogCrit(COMPONENT_FSAL,
@@ -1513,6 +1627,11 @@ static void release(struct fsal_obj_handle *obj_hdl)
 
 		handle_to_key(obj_hdl, &key);
 		vfs_state_release(&key);
+	} else if (type == DIRECTORY) {
+		if (myself->u.directory.path != NULL)
+			gsh_free(myself->u.directory.path);
+		if (myself->u.directory.fs_location != NULL)
+			gsh_free(myself->u.directory.fs_location);
 	} else if (vfs_unopenable_type(type)) {
 		if (myself->u.unopenable.name != NULL)
 			gsh_free(myself->u.unopenable.name);
@@ -1527,6 +1646,65 @@ static void release(struct fsal_obj_handle *obj_hdl)
 	gsh_free(myself);
 }
 
+/* vfs_fs_locations
+ * returns the saved referral information to NFS protocol layer
+ */
+
+static fsal_status_t vfs_fs_locations(struct fsal_obj_handle *obj_hdl,
+					struct fs_locations4 *fs_locs)
+{
+	struct vfs_fsal_obj_handle *myself;
+
+	myself = container_of(obj_hdl, struct vfs_fsal_obj_handle, obj_handle);
+	struct vfs_filesystem *vfs_fs = myself->obj_handle.fs->private_data;
+
+	LogFullDebug(COMPONENT_FSAL,
+		     "vfs_fs = %s root_fd = %d major = %d minor = %d",
+		     vfs_fs->fs->path, vfs_fs->root_fd,
+		     (int)vfs_fs->fs->fsid.major,
+		     (int)vfs_fs->fs->fsid.minor);
+
+	LogDebug(COMPONENT_FSAL,
+		 "fs_location = %p:%s",
+		 myself->u.directory.fs_location,
+		 myself->u.directory.fs_location);
+	if (myself->u.directory.fs_location != NULL) {
+		char *server;
+		char *path_sav, *path_work;
+
+		path_sav = gsh_strdup(myself->u.directory.fs_location);
+		path_work = path_sav;
+		server = strsep(&path_work, ":");
+
+		LogDebug(COMPONENT_FSAL,
+			 "fs_location server %s",
+			 server);
+		LogDebug(COMPONENT_FSAL,
+			 "fs_location path %s",
+			 path_work);
+
+		nfs4_pathname4_free(&fs_locs->fs_root);
+		nfs4_pathname4_alloc(&fs_locs->fs_root,
+				     myself->u.directory.path);
+		strncpy(fs_locs->locations.locations_val->
+			server.server_val->utf8string_val,
+			server, strlen(server));
+		fs_locs->locations.locations_val->
+			server.server_val->utf8string_len = strlen(server);
+		nfs4_pathname4_free(&fs_locs->
+				    locations.locations_val->rootpath);
+		nfs4_pathname4_alloc(&fs_locs->
+				     locations.locations_val->rootpath,
+				     path_work);
+
+		gsh_free(path_sav);
+	} else {
+		return fsalstat(ERR_FSAL_NOTSUPP, -1);
+	}
+
+	return fsalstat(ERR_FSAL_NO_ERROR, 0);
+}
+
 void vfs_handle_ops_init(struct fsal_obj_ops *ops)
 {
 	ops->release = release;
@@ -1541,8 +1719,9 @@ void vfs_handle_ops_init(struct fsal_obj_ops *ops)
 	ops->link = linkfile;
 	ops->rename = renamefile;
 	ops->unlink = file_unlink;
+	ops->fs_locations = vfs_fs_locations;
 	ops->close = vfs_close;
-	ops->handle_digest = handle_digest;
+	ops->handle_to_wire = handle_to_wire;
 	ops->handle_to_key = handle_to_key;
 	ops->open2 = vfs_open2;
 	ops->reopen2 = vfs_reopen2;
@@ -1647,10 +1826,7 @@ fsal_status_t vfs_lookup_path(struct fsal_export *exp_hdl,
 	close(dir_fd);
 
 	if (attrs_out != NULL) {
-		posix2fsal_attributes(&stat, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
+		posix2fsal_attributes_all(&stat, attrs_out);
 	}
 
 	*handle = &hdl->obj_handle;
@@ -1691,7 +1867,7 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 		if (*fs == NULL) {
 			LogInfo(COMPONENT_FSAL,
 				"Could not map fsid=0x%016"PRIx64
-				".0x%016"PRIx64" to filesytem",
+				".0x%016"PRIx64" to filesystem",
 				fsid.major, fsid.minor);
 			retval = ESTALE;
 			fsal_error = posix2fsal_error(retval);
@@ -1801,10 +1977,7 @@ fsal_status_t vfs_create_handle(struct fsal_export *exp_hdl,
 	}
 
 	if (attrs_out != NULL) {
-		posix2fsal_attributes(&obj_stat, attrs_out);
-
-		/* Make sure ATTR_RDATTR_ERR is cleared on success. */
-		attrs_out->mask &= ~ATTR_RDATTR_ERR;
+		posix2fsal_attributes_all(&obj_stat, attrs_out);
 	}
 
 	*handle = &hdl->obj_handle;
