@@ -1,7 +1,7 @@
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
- * Copyright 2015-2016 Red Hat, Inc. and/or its affiliates.
+ * Copyright 2015-2017 Red Hat, Inc. and/or its affiliates.
  * Author: Daniel Gryniewicz <dang@redhat.com>
  *
  * This program is free software; you can redistribute it and/or
@@ -97,10 +97,10 @@ static void mdcache_unexport(struct fsal_export *exp_hdl,
 	fsal_status_t status;
 	bool rc;
 
-	/* First unexport for the sub-FSAL */
-	subcall_raw(exp,
-		sub_export->exp_ops.unexport(sub_export, root_entry->sub_handle)
-	);
+	/* Indicate this export is going away so we don't create any new
+	 * export map entries.
+	 */
+	atomic_set_uint8_t_bits(&exp->flags, MDC_UNEXPORT);
 
 	/* Next, clean up our cache entries on the export */
 	while (true) {
@@ -124,7 +124,6 @@ static void mdcache_unexport(struct fsal_export *exp_hdl,
 			continue;
 		}
 
-
 		/* Must get attr_lock before mdc_exp_lock */
 		PTHREAD_RWLOCK_wrlock(&entry->attr_lock);
 		PTHREAD_RWLOCK_wrlock(&exp->mdc_exp_lock);
@@ -135,8 +134,6 @@ static void mdcache_unexport(struct fsal_export *exp_hdl,
 					   struct entry_export_map,
 					   export_per_entry);
 		if (expmap == NULL) {
-			/* Clear out first export pointer */
-			atomic_store_voidptr(&entry->first_export, NULL);
 			/* We must not hold entry->attr_lock across
 			 * try_cleanup_push (LRU lane lock order) */
 			PTHREAD_RWLOCK_unlock(&exp->mdc_exp_lock);
@@ -147,8 +144,9 @@ static void mdcache_unexport(struct fsal_export *exp_hdl,
 			mdcache_lru_cleanup_try_push(entry);
 		} else {
 			/* Make sure first export pointer is still valid */
-			atomic_store_voidptr(&entry->first_export,
-					     expmap->export);
+			atomic_store_int32_t(
+				&entry->first_export_id,
+				(int32_t) expmap->export->export.export_id);
 
 			PTHREAD_RWLOCK_unlock(&exp->mdc_exp_lock);
 			PTHREAD_RWLOCK_unlock(&entry->attr_lock);
@@ -158,6 +156,11 @@ static void mdcache_unexport(struct fsal_export *exp_hdl,
 		mdcache_put(entry);
 	};
 
+	/* Last unexport for the sub-FSAL */
+	subcall_raw(exp,
+		sub_export->exp_ops.unexport(sub_export, root_entry->sub_handle)
+	);
+
 	/* Unhash the root object */
 	rc = cih_remove_checked(root_entry);
 	assert(!rc);
@@ -704,18 +707,41 @@ static void mdcache_get_write_verifier(struct fsal_export *exp_hdl,
  * @param[in] flags	Related flages (currently endian)
  * @return FSAL status
  */
-static fsal_status_t mdcache_extract_handle(struct fsal_export *exp_hdl,
-					    fsal_digesttype_t in_type,
-					    struct gsh_buffdesc *fh_desc,
-					    int flags)
+static fsal_status_t mdcache_wire_to_host(struct fsal_export *exp_hdl,
+					  fsal_digesttype_t in_type,
+					  struct gsh_buffdesc *fh_desc,
+					  int flags)
 {
 	struct mdcache_fsal_export *exp = mdc_export(exp_hdl);
 	struct fsal_export *sub_export = exp->export.sub_export;
 	fsal_status_t status;
 
 	subcall_raw(exp,
-		status = sub_export->exp_ops.extract_handle(sub_export, in_type,
-							    fh_desc, flags)
+		status = sub_export->exp_ops.wire_to_host(sub_export, in_type,
+							  fh_desc, flags)
+	       );
+
+	return status;
+}
+
+/**
+ * @brief Produce handle-key from host-handle
+ *
+ * delegated to the underlying FSAL.
+ *
+ * @param[in] exp_hdl	Export to operate on
+ * @param[in,out] fh_desc	Source/dest for extracted digest
+ * @return FSAL status
+ */
+static fsal_status_t mdcache_host_to_key(struct fsal_export *exp_hdl,
+					    struct gsh_buffdesc *fh_desc)
+{
+	struct mdcache_fsal_export *exp = mdc_export(exp_hdl);
+	struct fsal_export *sub_export = exp->export.sub_export;
+	fsal_status_t status;
+
+	subcall_raw(exp,
+		status = sub_export->exp_ops.host_to_key(sub_export, fh_desc)
 	       );
 
 	return status;
@@ -765,6 +791,29 @@ static void mdcache_free_state(struct fsal_export *exp_hdl,
 	       );
 }
 
+/**
+ * @brief Check to see if a user is superuser
+ *
+ * @param[in] exp_hdl               Export state_t is associated with
+ * @param[in] creds                 Credentials to check for superuser
+ *
+ * @returns NULL on failure otherwise a state structure.
+ */
+
+static bool mdcache_is_superuser(struct fsal_export *exp_hdl,
+				 const struct user_cred *creds)
+{
+	struct mdcache_fsal_export *exp = mdc_export(exp_hdl);
+	struct fsal_export *sub_export = exp->export.sub_export;
+	bool status;
+
+	subcall_raw(exp,
+		status = sub_export->exp_ops.is_superuser(sub_export, creds)
+	       );
+
+	return status;
+}
+
 /* mdcache_export_ops_init
  * overwrite vector entries with the methods that we support
  */
@@ -776,7 +825,8 @@ void mdcache_export_ops_init(struct export_ops *ops)
 	ops->release = mdcache_exp_release;
 	ops->lookup_path = mdcache_lookup_path;
 	/* lookup_junction unimplemented because deprecated */
-	ops->extract_handle = mdcache_extract_handle;
+	ops->wire_to_host = mdcache_wire_to_host;
+	ops->host_to_key = mdcache_host_to_key;
 	ops->create_handle = mdcache_create_handle;
 	ops->get_fs_dynamic_info = mdcache_get_dynamic_info;
 	ops->fs_supports = mdcache_fs_supports;
@@ -802,6 +852,7 @@ void mdcache_export_ops_init(struct export_ops *ops)
 	ops->get_write_verifier = mdcache_get_write_verifier;
 	ops->alloc_state = mdcache_alloc_state;
 	ops->free_state = mdcache_free_state;
+	ops->is_superuser = mdcache_is_superuser;
 }
 
 #if 0
@@ -851,6 +902,7 @@ mdc_init_export(struct fsal_module *fsal_hdl,
 	struct mdcache_fsal_export *myself;
 	int namelen;
 	pthread_rwlockattr_t attrs;
+	fsal_status_t status = {0, 0};
 
 	myself = gsh_calloc(1, sizeof(struct mdcache_fsal_export));
 	namelen = strlen(op_ctx->fsal_export->fsal->name) + 5;
@@ -864,10 +916,15 @@ mdc_init_export(struct fsal_module *fsal_hdl,
 	mdcache_export_ops_init(&myself->export.exp_ops);
 	myself->super_up_ops = *super_up_ops; /* Struct copy */
 	myself->up_ops = *mdc_up_ops; /* Struct copy */
-	myself->up_ops.up_export = &myself->export;
+	myself->up_ops.up_gsh_export = op_ctx->ctx_export;
+	myself->up_ops.up_fsal_export = &myself->export;
 	myself->export.up_ops = &myself->up_ops;
 	myself->export.fsal = fsal_hdl;
-	fsal_attach_export(fsal_hdl, &myself->export.exports);
+	status.minor = fsal_attach_export(fsal_hdl, &myself->export.exports);
+	if (status.minor != 0) {
+		status.major = posix2fsal_error(status.minor);
+		return status;
+	}
 	fsal_export_stack(op_ctx->fsal_export, &myself->export);
 
 	glist_init(&myself->entry_list);
