@@ -77,7 +77,6 @@ enum config_type {
 	CONFIG_BOOL,
 	CONFIG_BOOLBIT,
 	CONFIG_IP_ADDR,
-	CONFIG_INET_PORT,
 	CONFIG_BLOCK,
 	CONFIG_PROC
 };
@@ -320,6 +319,8 @@ struct config_item {
 			uint64_t minval;
 			uint64_t maxval;
 			uint64_t def;
+			uint32_t bit;
+			size_t set_off;
 		} ui64;
 		struct { /* CONFIG_FSID */
 			int64_t def_maj;
@@ -638,25 +639,6 @@ struct config_item {
 	  .off = offsetof(struct _struct_, _mem_)   \
 	}
 
-#define CONF_ITEM_INET_PORT(_name_, _min_, _max_, _def_, _struct_, _mem_) \
-	{ .name = _name_,			    \
-	  .type = CONFIG_INET_PORT,		    \
-	  .u.i16.minval = _min_,		    \
-	  .u.i16.maxval = _max_,		    \
-	  .u.i16.def = _def_,			    \
-	  .off = offsetof(struct _struct_, _mem_)   \
-	}
-
-#define CONF_MAND_INET_PORT(_name_, _min_, _max_, _def_, _struct_, _mem_) \
-	{ .name = _name_,			    \
-	  .type = CONFIG_INET_PORT,		    \
-	  .flags = CONFIG_UNIQUE|CONFIG_MANDATORY,  \
-	  .u.i16.minval = _min_,		    \
-	  .u.i16.maxval = _max_,		    \
-	  .u.i16.def = _def_,			    \
-	  .off = offsetof(struct _struct_, _mem_)   \
-	}
-
 #define CONF_ITEM_I16(_name_, _min_, _max_, _def_, _struct_, _mem_) \
 	{ .name = _name_,			    \
 	  .type = CONFIG_INT16,		    \
@@ -767,6 +749,19 @@ struct config_item {
 	  .off = offsetof(struct _struct_, _mem_)   \
 	}
 
+#define CONF_ITEM_UI64_SET(_name_, _min_, _max_, _def_, _struct_, \
+			  _mem_, _bit_, _set_)       \
+	{ .name = _name_,			    \
+	  .type = CONFIG_UINT64,		    \
+	  .flags = CONFIG_MARK_SET,		    \
+	  .u.ui64.minval = _min_,		    \
+	  .u.ui64.maxval = _max_,		    \
+	  .u.ui64.def = _def_,			    \
+	  .u.ui64.bit = _bit_,			    \
+	  .u.ui64.set_off = offsetof(struct _struct_, _set_),   \
+	  .off = offsetof(struct _struct_, _mem_)   \
+	}
+
 #define CONFIG_EOL {.name = NULL, .type = CONFIG_NULL}
 
 /**
