@@ -233,7 +233,7 @@ void dec_nlm_state_ref(state_t *state);
 int get_nlm_state(enum state_type state_type,
 		  struct fsal_obj_handle *state_obj,
 		  state_owner_t *state_owner,
-		  bool nsm_state_applies,
+		  care_t care,
 		  uint32_t nsm_state,
 		  state_t **pstate);
 
@@ -575,7 +575,7 @@ static inline bool owner_has_state(state_owner_t *owner)
 	/* If the owner is on the cached owners list, there can't be
 	 * active state.
 	 */
-	if (atomic_fetch_time_t(&nfs4_owner->cache_expire) != 0)
+	if (atomic_fetch_time_t(&nfs4_owner->so_cache_expire) != 0)
 		return false;
 
 	PTHREAD_MUTEX_lock(&owner->so_mutex);
@@ -696,31 +696,65 @@ static inline void *convert_lock_owner(struct fsal_export *fsal_export,
  *
  ******************************************************************************/
 
-state_status_t state_add_impl(struct fsal_obj_handle *obj,
-			      enum state_type state_type,
-			      union state_data *state_data,
-			      state_owner_t *owner_input, state_t **state,
-			      struct state_refer *refer);
+#define state_add_impl(o, t, d, i, s, r) \
+	_state_add_impl(o, t, d, i, s, r, __func__, __LINE__)
+state_status_t _state_add_impl(struct fsal_obj_handle *obj,
+			       enum state_type state_type,
+			       union state_data *state_data,
+			       state_owner_t *owner_input, state_t **state,
+			       struct state_refer *refer,
+			       const char *func, int line);
 
-state_status_t state_add(struct fsal_obj_handle *obj,
-			 enum state_type state_type,
-			 union state_data *state_data,
-			 state_owner_t *owner_input,
-			 state_t **state, struct state_refer *refer);
+#define state_add(o, t, d, i, s, r) \
+	_state_add(o, t, d, i, s, r, __func__, __LINE__)
+state_status_t _state_add(struct fsal_obj_handle *obj,
+			  enum state_type state_type,
+			  union state_data *state_data,
+			  state_owner_t *owner_input,
+			  state_t **state, struct state_refer *refer,
+			  const char *func, int line);
 
 state_status_t state_set(state_t *state);
 
-void state_del_locked(state_t *state);
+#define state_del_locked(s) _state_del_locked(s, __func__, __LINE__)
+void _state_del_locked(state_t *state, const char *func, int line);
 
 void state_del(state_t *state);
 
+/**
+ * @brief Get a reference to the obj owning a state
+ *
+ * @note state_mutex MUST be held
+ *
+ * @param[in] state	State to get from
+ * @return obj handle on success
+ */
+static inline struct fsal_obj_handle *get_state_obj_ref_locked(state_t *state)
+{
+	if (state->state_obj) {
+		state->state_obj->obj_ops.get_ref(state->state_obj);
+	}
+
+	return state->state_obj;
+}
+
+/**
+ * @brief Get a reference to the obj owning a state
+ *
+ * Takes state_mutex, so it should not be held.
+ *
+ * @param[in] state	State to get from
+ * @return obj handle on success
+ */
 static inline struct fsal_obj_handle *get_state_obj_ref(state_t *state)
 {
-	if (!state->state_obj)
-		return NULL;
+	struct fsal_obj_handle *obj;
 
-	state->state_obj->obj_ops.get_ref(state->state_obj);
-	return state->state_obj;
+	PTHREAD_MUTEX_lock(&state->state_mutex);
+	obj = get_state_obj_ref_locked(state);
+	PTHREAD_MUTEX_unlock(&state->state_mutex);
+
+	return obj;
 }
 
 static inline struct gsh_export *get_state_export_ref(state_t *state)
@@ -989,8 +1023,6 @@ void blocked_lock_polling(struct fridgethr_context *ctx);
 
 void nfs4_start_grace(nfs_grace_start_t *gsp);
 int nfs_in_grace(void);
-void nfs4_create_clid_name(nfs_client_record_t *, nfs_client_id_t *,
-			   struct svc_req *);
 void nfs4_add_clid(nfs_client_id_t *);
 void nfs4_rm_clid(const char *, char *, int);
 void nfs4_chk_clid(nfs_client_id_t *);
@@ -1000,6 +1032,28 @@ void nfs4_create_recov_dir(void);
 void nfs4_record_revoke(nfs_client_id_t *, nfs_fh4 *);
 bool nfs4_check_deleg_reclaim(nfs_client_id_t *, nfs_fh4 *);
 
+/**
+ * @brief Check to see if an object is a junction
+ *
+ * @param[in] obj	Object to check
+ * @return true if junction, false otherwise
+ */
+static inline bool obj_is_junction(struct fsal_obj_handle *obj)
+{
+	bool res = false;
+
+	if (obj->type != DIRECTORY)
+		return false;
+
+	PTHREAD_RWLOCK_rdlock(&obj->state_hdl->state_lock);
+
+	if ((obj->state_hdl->dir.junction_export != NULL ||
+	     atomic_fetch_int32_t(&obj->state_hdl->dir.exp_root_refcount) != 0))
+		res = true;
+	PTHREAD_RWLOCK_unlock(&obj->state_hdl->state_lock);
+
+	return res;
+}
 
 #endif				/* SAL_FUNCTIONS_H */
 
