@@ -96,7 +96,7 @@ static int StrExportOptions(struct display_buffer *dspbuf,
 	if (b_left <= 0)
 		return b_left;
 
-	b_left = display_printf(dspbuf, "options=%08 "PRIx32, p_perms->options);
+	b_left = display_printf(dspbuf, "options=%08"PRIx32, p_perms->options);
 
 	if (b_left <= 0)
 		return b_left;
@@ -315,7 +315,7 @@ void LogClientListEntry(log_levels_t level,
 			char *tag,
 			exportlist_client_entry_t *entry)
 {
-	char perms[1024];
+	char perms[1024] = "\0";
 	struct display_buffer dspbuf = {sizeof(perms), perms, perms};
 	char addr[INET6_ADDRSTRLEN];
 	char *paddr = addr;
@@ -488,13 +488,27 @@ static int add_client(struct glist_head *client_list,
 	case TERM_V4ADDR:
 		rc = inet_pton(AF_INET, client_tok,
 			       &cli->client.hostif.clientaddr);
-		assert(rc == 1);  /* this had better be grok'd by now! */
+		if (rc != 1) {
+			config_proc_error(cnode, err_type,
+					  "IPv4 addr (%s) not in presentation format",
+					  client_tok);
+			err_type->invalid = true;
+			errcnt++;
+			goto out;
+		}
 		cli->type = HOSTIF_CLIENT;
 		break;
 	case TERM_V6ADDR:
 		rc = inet_pton(AF_INET6, client_tok,
 			       &cli->client.hostif.clientaddr6);
-		assert(rc == 1);  /* this had better be grok'd by now! */
+		if (rc != 1) {
+			config_proc_error(cnode, err_type,
+					  "IPv6 addr (%s) not in presentation format",
+					  client_tok);
+			err_type->invalid = true;
+			errcnt++;
+			goto out;
+		}
 		cli->type = HOSTIF_CLIENT_V6;
 		break;
 	case TERM_TOKEN: /* only dns names now. */
@@ -991,7 +1005,7 @@ static int export_commit_common(void *node, void *link_mem, void *self_struct,
 {
 	struct gsh_export *export = self_struct, *probe_exp;
 	int errcnt = 0;
-	char perms[1024];
+	char perms[1024] = "\0";
 	struct display_buffer dspbuf = {sizeof(perms), perms, perms};
 
 	LogFullDebug(COMPONENT_EXPORT, "Processing %p", export);
@@ -1000,7 +1014,7 @@ static int export_commit_common(void *node, void *link_mem, void *self_struct,
 	if (export->export_perms.options & EXPORT_OPTION_NFSV4) {
 		if (export->pseudopath == NULL) {
 			LogCrit(COMPONENT_CONFIG,
-				"Exporting to NFSv4 but not Pseudo path defined");
+				"Exporting to NFSv4 but no Pseudo path defined");
 			err_type->invalid = true;
 			errcnt++;
 			return errcnt;
@@ -1014,6 +1028,18 @@ static int export_commit_common(void *node, void *link_mem, void *self_struct,
 			return errcnt;
 		}
 	}
+
+	/* If we are using mount_path_pseudo = true we MUST have a Pseudo Path.
+	 */
+	if (nfs_param.core_param.mount_path_pseudo &&
+	    export->pseudopath == NULL) {
+		LogCrit(COMPONENT_CONFIG,
+			"NFS_CORE_PARAM mount_path_pseudo is TRUE but no Pseudo path defined");
+		err_type->invalid = true;
+		errcnt++;
+		return errcnt;
+	}
+
 	if (export->pseudopath != NULL &&
 	    export->pseudopath[0] != '/') {
 		LogCrit(COMPONENT_CONFIG,
@@ -1036,7 +1062,7 @@ static int export_commit_common(void *node, void *link_mem, void *self_struct,
 		if ((export->export_perms.options &
 		     EXPORT_OPTION_PROTOCOLS) != EXPORT_OPTION_NFSV4) {
 			LogCrit(COMPONENT_CONFIG,
-				"Export id 0 must indicate Protocols=4");
+				"Export id 0 must include 4 in Protocols");
 			err_type->invalid = true;
 			errcnt++;
 		}
@@ -1339,7 +1365,7 @@ static void export_display(const char *step, void *node,
 			   void *link_mem, void *self_struct)
 {
 	struct gsh_export *export = self_struct;
-	char perms[1024];
+	char perms[1024] = "\0";
 	struct display_buffer dspbuf = {sizeof(perms), perms, perms};
 
 	(void) StrExportOptions(&dspbuf, &export->export_perms);
@@ -1400,7 +1426,7 @@ static int export_defaults_commit(void *node, void *link_mem,
 				  void *self_struct,
 				  struct config_error_type *err_type)
 {
-	char perms[1024];
+	char perms[1024] = "\0";
 	struct display_buffer dspbuf = {sizeof(perms), perms, perms};
 
 	(void) StrExportOptions(&dspbuf, &export_opt_cfg.conf);
@@ -1426,7 +1452,7 @@ static void export_defaults_display(const char *step, void *node,
 				    void *link_mem, void *self_struct)
 {
 	struct export_perms *defaults = self_struct;
-	char perms[1024];
+	char perms[1024] = "\0";
 	struct display_buffer dspbuf = {sizeof(perms), perms, perms};
 
 	(void) StrExportOptions(&dspbuf, defaults);
@@ -1669,14 +1695,18 @@ static struct config_item fsal_params[] = {
 		       _struct_, fullpath), /* must chomp '/' */	\
 	CONF_UNIQ_PATH("Pseudo", 1, MAXPATHLEN, NULL,			\
 		       _struct_, pseudopath),				\
-	CONF_ITEM_UI64("MaxRead", 512, FSAL_MAXIOSIZE, FSAL_MAXIOSIZE,	\
-		       _struct_, MaxRead),				\
-	CONF_ITEM_UI64("MaxWrite", 512, FSAL_MAXIOSIZE, FSAL_MAXIOSIZE,	\
-		       _struct_, MaxWrite),				\
-	CONF_ITEM_UI64("PrefRead", 512, FSAL_MAXIOSIZE, FSAL_MAXIOSIZE,	\
-		       _struct_, PrefRead),				\
-	CONF_ITEM_UI64("PrefWrite", 512, FSAL_MAXIOSIZE, FSAL_MAXIOSIZE,\
-		       _struct_, PrefWrite),				\
+	CONF_ITEM_UI64_SET("MaxRead", 512, FSAL_MAXIOSIZE,		\
+			FSAL_MAXIOSIZE, _struct_, MaxRead,		\
+			EXPORT_OPTION_MAXREAD_SET, options_set),	\
+	CONF_ITEM_UI64_SET("MaxWrite", 512, FSAL_MAXIOSIZE,		\
+			FSAL_MAXIOSIZE, _struct_, MaxWrite,		\
+			EXPORT_OPTION_MAXWRITE_SET, options_set),	\
+	CONF_ITEM_UI64_SET("PrefRead", 512, FSAL_MAXIOSIZE,		\
+			FSAL_MAXIOSIZE, _struct_, PrefRead,		\
+			EXPORT_OPTION_PREFREAD_SET, options_set),	\
+	CONF_ITEM_UI64_SET("PrefWrite", 512, FSAL_MAXIOSIZE,		\
+			FSAL_MAXIOSIZE, _struct_, PrefWrite,		\
+			EXPORT_OPTION_PREFWRITE_SET, options_set),	\
 	CONF_ITEM_UI64("PrefReaddir", 512, FSAL_MAXIOSIZE, 16384,	\
 		       _struct_, PrefReaddir),				\
 	CONF_ITEM_FSID_SET("Filesystem_id", 666, 666,			\
@@ -1884,7 +1914,11 @@ static int build_default_root(struct config_error_type *err_type)
 
 	export->options = EXPORT_OPTION_USE_COOKIE_VERIFIER;
 	export->options_set = EXPORT_OPTION_FSID_SET |
-			      EXPORT_OPTION_USE_COOKIE_VERIFIER;
+			      EXPORT_OPTION_USE_COOKIE_VERIFIER |
+			      EXPORT_OPTION_MAXREAD_SET |
+			      EXPORT_OPTION_MAXWRITE_SET |
+			      EXPORT_OPTION_PREFREAD_SET |
+			      EXPORT_OPTION_PREFWRITE_SET;
 
 	/* Set the fullpath to "/" */
 	export->fullpath = gsh_strdup("/");
@@ -2088,7 +2122,14 @@ void free_export_resources(struct gsh_export *export)
 
 static bool init_export_cb(struct gsh_export *exp, void *state)
 {
-	return !(init_export_root(exp));
+	struct glist_head *errlist = state;
+
+	if (init_export_root(exp)) {
+		glist_del(&exp->exp_list);
+		glist_add(errlist, &exp->exp_list);
+	}
+
+	return true;
 }
 
 /**
@@ -2097,7 +2138,17 @@ static bool init_export_cb(struct gsh_export *exp, void *state)
 
 void exports_pkginit(void)
 {
-	foreach_gsh_export(init_export_cb, NULL);
+	struct glist_head errlist;
+	struct glist_head *glist, *glistn;
+	struct gsh_export *export;
+
+	glist_init(&errlist);
+	foreach_gsh_export(init_export_cb, true, &errlist);
+
+	glist_for_each_safe(glist, glistn, &errlist) {
+		export = glist_entry(glist, struct gsh_export, exp_list);
+		export_revert(export);
+	}
 }
 
 /**
@@ -2135,6 +2186,52 @@ fsal_status_t nfs_export_get_root_entry(struct gsh_export *export,
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
+/**
+ * @brief Set file systems max read write sizes in the export
+ *
+ * @param export [IN] the export
+ * @param maxread [IN] maxread size
+ * @param maxwrite [IN] maxwrite size
+ */
+
+static void set_fs_max_rdwr_size(struct gsh_export *export, uint64_t maxread,
+				 uint64_t maxwrite)
+{
+	if (maxread != 0) {
+		if (!op_ctx_export_has_option_set(EXPORT_OPTION_MAXREAD_SET)) {
+			LogInfo(COMPONENT_EXPORT,
+				"Readjusting MaxRead to %" PRIu64,
+				maxread);
+			export->MaxRead = maxread;
+		}
+
+		if (!op_ctx_export_has_option_set(EXPORT_OPTION_PREFREAD_SET) ||
+		    (export->PrefRead > export->MaxRead)) {
+			LogInfo(COMPONENT_EXPORT,
+				"Readjusting PrefRead to %"PRIu64,
+				export->MaxRead);
+			export->PrefRead = export->MaxRead;
+		}
+	}
+
+	if (maxwrite != 0) {
+		if (!op_ctx_export_has_option_set(EXPORT_OPTION_MAXWRITE_SET)) {
+			LogInfo(COMPONENT_EXPORT,
+				"Readjusting MaxWrite to %"PRIu64,
+				maxwrite);
+			export->MaxWrite = maxwrite;
+		}
+
+		if (!op_ctx_export_has_option_set(EXPORT_OPTION_PREFWRITE_SET)
+		    || (export->PrefWrite > export->MaxWrite)) {
+			LogInfo(COMPONENT_EXPORT,
+				"Readjusting PrefWrite to %"PRIu64,
+				export->MaxWrite);
+			export->PrefWrite = export->MaxWrite;
+		}
+	}
+}
+
 /**
  * @brief Initialize the root cache inode for an export.
  *
@@ -2177,6 +2274,26 @@ int init_export_root(struct gsh_export *export)
 		goto out;
 	}
 
+	if (!op_ctx_export_has_option_set(EXPORT_OPTION_MAXREAD_SET) ||
+	    !op_ctx_export_has_option_set(EXPORT_OPTION_MAXWRITE_SET) ||
+	    !op_ctx_export_has_option_set(EXPORT_OPTION_PREFREAD_SET) ||
+	    !op_ctx_export_has_option_set(EXPORT_OPTION_PREFWRITE_SET)) {
+
+		fsal_dynamicfsinfo_t dynamicinfo;
+
+		dynamicinfo.maxread = 0;
+		dynamicinfo.maxwrite = 0;
+		fsal_status =
+			export->fsal_export->exp_ops.get_fs_dynamic_info(
+				export->fsal_export, obj, &dynamicinfo);
+
+		if (!FSAL_IS_ERROR(fsal_status)) {
+			set_fs_max_rdwr_size(export,
+					     dynamicinfo.maxread,
+					     dynamicinfo.maxwrite);
+		}
+	}
+
 	PTHREAD_RWLOCK_wrlock(&obj->state_hdl->state_lock);
 	PTHREAD_RWLOCK_wrlock(&export->lock);
 
@@ -2207,20 +2324,6 @@ out:
 	return my_status;
 }
 
-static inline void clean_up_export(struct gsh_export *export,
-				   struct fsal_obj_handle *root_obj)
-{
-	/* Make export unreachable */
-	pseudo_unmount_export(export);
-	remove_gsh_export(export->export_id);
-
-	/* Release state belonging to this export */
-	state_release_export(export);
-
-	/* Flush FSAL-specific state */
-	export->fsal_export->exp_ops.unexport(export->fsal_export, root_obj);
-}
-
 /**
  * @brief Release all the export state, including the root object
  *
@@ -2262,7 +2365,21 @@ static void release_export(struct gsh_export *export)
 		 "Released root obj %p for path %s on export_id=%d",
 		 obj, export->fullpath, export->export_id);
 
-	clean_up_export(export, obj);
+	/* Make export unreachable via pseudo fs.
+	 * We keep the export in the export hash table through the following
+	 * so that the underlying FSALs have access to the export while
+	 * performing the various cleanup operations.
+	 */
+	pseudo_unmount_export(export);
+
+	/* Release state belonging to this export */
+	state_release_export(export);
+
+	/* Flush FSAL-specific state */
+	export->fsal_export->exp_ops.unexport(export->fsal_export, obj);
+
+	/* Remove the mapping to the export now that cleanup is complete. */
+	remove_gsh_export(export->export_id);
 
 	/* Release ref taken above */
 	obj->obj_ops.put_ref(obj);
@@ -2506,13 +2623,13 @@ static exportlist_client_entry_t *client_match_any(sockaddr_t *hostaddr,
  */
 bool export_check_security(struct svc_req *req)
 {
-	switch (req->rq_cred.oa_flavor) {
+	switch (req->rq_msg.cb_cred.oa_flavor) {
 	case AUTH_NONE:
 		if ((op_ctx->export_perms->options &
 		     EXPORT_OPTION_AUTH_NONE) == 0) {
 			LogInfo(COMPONENT_EXPORT,
 				"Export %s does not support AUTH_NONE",
-				op_ctx->ctx_export->fullpath);
+				op_ctx_export_path(op_ctx->ctx_export));
 			return false;
 		}
 		break;
@@ -2522,7 +2639,7 @@ bool export_check_security(struct svc_req *req)
 		     EXPORT_OPTION_AUTH_UNIX) == 0) {
 			LogInfo(COMPONENT_EXPORT,
 				"Export %s does not support AUTH_UNIX",
-				op_ctx->ctx_export->fullpath);
+				op_ctx_export_path(op_ctx->ctx_export));
 			return false;
 		}
 		break;
@@ -2535,14 +2652,13 @@ bool export_check_security(struct svc_req *req)
 				 EXPORT_OPTION_RPCSEC_GSS_PRIV)) == 0) {
 			LogInfo(COMPONENT_EXPORT,
 				"Export %s does not support RPCSEC_GSS",
-				op_ctx->ctx_export->fullpath);
+				op_ctx_export_path(op_ctx->ctx_export));
 			return false;
 		} else {
-			struct svc_rpc_gss_data *gd;
-			rpc_gss_svc_t svc;
+			struct rpc_gss_cred *gc = (struct rpc_gss_cred *)
+				req->rq_msg.rq_cred_body;
+			rpc_gss_svc_t svc = gc->gc_svc;
 
-			gd = SVCAUTH_PRIVATE(req->rq_auth);
-			svc = gd->sec.svc;
 			LogFullDebug(COMPONENT_EXPORT, "Testing svc %d",
 				     (int)svc);
 			switch (svc) {
@@ -2551,7 +2667,8 @@ bool export_check_security(struct svc_req *req)
 				     EXPORT_OPTION_RPCSEC_GSS_NONE) == 0) {
 					LogInfo(COMPONENT_EXPORT,
 						"Export %s does not support RPCSEC_GSS_SVC_NONE",
-						op_ctx->ctx_export->fullpath);
+						op_ctx_export_path(
+							op_ctx->ctx_export));
 					return false;
 				}
 				break;
@@ -2561,7 +2678,8 @@ bool export_check_security(struct svc_req *req)
 				     EXPORT_OPTION_RPCSEC_GSS_INTG) == 0) {
 					LogInfo(COMPONENT_EXPORT,
 						"Export %s does not support RPCSEC_GSS_SVC_INTEGRITY",
-						op_ctx->ctx_export->fullpath);
+						op_ctx_export_path(
+							op_ctx->ctx_export));
 					return false;
 				}
 				break;
@@ -2571,7 +2689,8 @@ bool export_check_security(struct svc_req *req)
 				     EXPORT_OPTION_RPCSEC_GSS_PRIV) == 0) {
 					LogInfo(COMPONENT_EXPORT,
 						"Export %s does not support RPCSEC_GSS_SVC_PRIVACY",
-						op_ctx->ctx_export->fullpath);
+						op_ctx_export_path(
+							op_ctx->ctx_export));
 					return false;
 				}
 				break;
@@ -2579,7 +2698,7 @@ bool export_check_security(struct svc_req *req)
 			default:
 				LogInfo(COMPONENT_EXPORT,
 					"Export %s does not support unknown RPCSEC_GSS_SVC %d",
-					op_ctx->ctx_export->fullpath,
+					op_ctx_export_path(op_ctx->ctx_export),
 					(int)svc);
 				return false;
 			}
@@ -2589,8 +2708,8 @@ bool export_check_security(struct svc_req *req)
 	default:
 		LogInfo(COMPONENT_EXPORT,
 			"Export %s does not support unknown oa_flavor %d",
-			op_ctx->ctx_export->fullpath,
-			(int)req->rq_cred.oa_flavor);
+			op_ctx_export_path(op_ctx->ctx_export),
+			(int)req->rq_msg.cb_cred.oa_flavor);
 		return false;
 	}
 
@@ -2748,9 +2867,9 @@ void export_check_access(void)
 		(void) sprint_sockip(hostaddr,
 				     ipstring, sizeof(ipstring));
 		LogMidDebug(COMPONENT_EXPORT,
-			    "Check for address %s for export id %u fullpath %s",
+			    "Check for address %s for export id %u path %s",
 			    ipstring, op_ctx->ctx_export->export_id,
-			    op_ctx->ctx_export->fullpath);
+			    op_ctx_export_path(op_ctx->ctx_export));
 	}
 
 	/* Does the client match anyone on the client list? */
@@ -2829,7 +2948,7 @@ void export_check_access(void)
 	op_ctx->export_perms->set |= export_opt.def.set;
 
 	if (isMidDebug(COMPONENT_EXPORT)) {
-		char perms[1024];
+		char perms[1024] = "\0";
 		struct display_buffer dspbuf = {sizeof(perms), perms, perms};
 
 		if (client != NULL) {
